import uuid

from .planner import RoutePlanner


class _Config:
    __slots__ = [
        "HOST",
        "PORT",
        "PASSWORD",
        "HANDSHAKE_INTERVAL",
        "HANDSHAKE_TIMEOUT",
        "IPBLOCKS",
        "EXCLUDEIPS",
        "_RoutePlanner",
        "DEFAULT_AUTOPLAY",
        "DEFAULT_VOLUME",
        "DEFAULT_CROSSFADE",
        "DEFAULT_GAPLESS",
        "SAMPLING_RATE",
        "CHANNELS",
        "FRAME_LENGTH",
        "SAMPLE_SIZE",
        "EXPECTED_PACKETLOSS",
        "BITRATE",
        "BUFFERLIMIT",
        "PRELOAD_TIME",
        "VCTIMEOUT",
        "ENABLED_EXT_EXTRACTOR",
        "SPOTIFY_ID",
        "SPOTIFY_SECRET",
        "RANDOM_STATE",
    ]

    def __init__(self) -> None:
        # SERVER
        self.HOST = "0.0.0.0"
        self.PORT = 8000
        self.PASSWORD = "hellodiscodo"
        self.HANDSHAKE_INTERVAL = 15
        self.HANDSHAKE_TIMEOUT = 60.0

        # NETWORK
        self.IPBLOCKS = []
        self.EXCLUDEIPS = []

        # PLAYER
        self.DEFAULT_AUTOPLAY = True
        self.DEFAULT_VOLUME = 1.0
        self.DEFAULT_CROSSFADE = 10.0
        self.DEFAULT_GAPLESS = False

        # AUDIO
        self.SAMPLING_RATE = 48000
        self.CHANNELS = 2
        self.FRAME_LENGTH = 20
        self.SAMPLE_SIZE = 4
        self.EXPECTED_PACKETLOSS = 0
        self.BITRATE = 128

        # BUFFER
        self.BUFFERLIMIT = 5
        self.PRELOAD_TIME = 10

        # CONNECTION
        self.VCTIMEOUT = 300.0

        # EXTRA EXTRACTOR
        self.ENABLED_EXT_EXTRACTOR = []  # melon, spotify

        self.SPOTIFY_ID = None
        self.SPOTIFY_SECRET = None

        # ETC
        self.RANDOM_STATE = str(uuid.uuid4())

    @property
    def SAMPLES_PER_FRAME(self) -> int:
        return int(self.SAMPLING_RATE / 1000 * self.FRAME_LENGTH)

    @property
    def FRAME_SIZE(self) -> int:
        return self.SAMPLES_PER_FRAME ** 2

    @property
    def DELAY(self) -> float:
        return self.FRAME_LENGTH / 1000.0

    @property
    def RoutePlanner(self) -> RoutePlanner:
        if not self.IPBLOCKS:
            return None

        if not hasattr(self, "_RoutePlanner"):
            self._RoutePlanner = RoutePlanner(self.IPBLOCKS, self.EXCLUDEIPS)

        return self._RoutePlanner

    def from_dict(self, data: dict) -> None:
        for Key, Value in data.items():
            setattr(self, Key, Value)


Config = _Config()



class DiscodoException(Exception):
    pass


class Forbidden(DiscodoException):
    pass


class TooManyRequests(DiscodoException):
    pass


class NoSearchResults(DiscodoException):
    pass


class VoiceClientNotFound(DiscodoException):
    pass


class NotConnected(DiscodoException):
    pass


class NotSeekable(DiscodoException):
    pass


class NotPlaying(DiscodoException):
    pass


class NodeNotConnected(DiscodoException):
    pass



import argparse
import asyncio
import json
import logging
import os
import sys

import colorlog

from . import __version__
from .config import Config
from .natives import opus

log = logging.getLogger("discodo")


if not opus.isLoaded() and not opus.loadDefaultOpus():
    raise ValueError(
        "Cannot load libopus, please install `libopus-dev` if you are using linux."
    )


class loggingFilter(logging.Filter):
    def __init__(self, level) -> None:
        self.level = level

    def filter(self, record) -> bool:
        return record.levelno < self.level


stdoutHandler = logging.StreamHandler(sys.stdout)
stderrHandler = logging.StreamHandler(sys.stderr)
stdoutHandler.addFilter(loggingFilter(logging.WARNING))


def setLoggingLevel(level) -> None:
    for logger in [log, logging.getLogger("libav")]:
        addLoggingHandler(logger)
        logger.setLevel(logging.DEBUG)
        stdoutHandler.setLevel(level)
        stderrHandler.setLevel(max(level, logging.WARNING))


ColoredFormatter = colorlog.ColoredFormatter(
    "[%(asctime)s] [%(filename)s] [%(name)s:%(bold)s%(module)s%(reset)s] %(log_color)s[%(levelname)s]%(reset)s: %(message)s",
    datefmt=None,
    reset=True,
    log_colors={
        "DEBUG": "green",
        "INFO": "cyan",
        "WARNING": "yellow",
        "ERROR": "red",
        "CRITICAL": "red,bg_white",
    },
    secondary_log_colors={},
    style="%",
)
stdoutHandler.setFormatter(ColoredFormatter)
stderrHandler.setFormatter(ColoredFormatter)


def addLoggingHandler(logger) -> None:
    logger.addHandler(stdoutHandler)
    logger.addHandler(stderrHandler)


parser = argparse.ArgumentParser("discodo")


def is_valid_file(parser, args):
    if not os.path.exists(args):
        parser.error(f"The file {args} does not exist!")

    with open(args, "r") as fp:
        return json.load(fp)


parser.add_argument(
    "--version",
    action="version",
    version=f"%(prog)s {__version__}",
    help="Config json file path (default: None)",
)

parser.add_argument(
    "--config",
    type=lambda args: is_valid_file(parser, args),
    default={},
    help="Config json file path (default: None)",
)

webGroup = parser.add_argument_group("Webserver Option")

webGroup.add_argument(
    "--host",
    "-H",
    type=str,
    default="0.0.0.0",
    help="the hostname to listen on (default: 0.0.0.0)",
)
webGroup.add_argument(
    "--port",
    "-P",
    type=int,
    default=8000,
    help="the port of the webserver (default: 8000)",
)
webGroup.add_argument(
    "--auth",
    "-A",
    type=str,
    default="hellodiscodo",
    help="the password of the webserver (default: hellodiscodo)",
)
webGroup.add_argument(
    "--ws-interval",
    type=int,
    default=15,
    help="heartbeat interval between discodo server and client (default: 15)",
)
webGroup.add_argument(
    "--ws-timeout",
    type=int,
    default=60,
    help="seconds to close connection there is no respond from client (default: 60)",
)

networkGroup = parser.add_argument_group("Network Option")
networkGroup.add_argument(
    "--ip",
    type=str,
    action="append",
    default=[],
    help="Client-side IP blocks to use",
)
networkGroup.add_argument(
    "--exclude-ip",
    type=str,
    action="append",
    default=[],
    help="Client-side IP addresses not to use",
)

playerGroup = parser.add_argument_group("Player Option")

playerGroup.add_argument(
    "--default-volume",
    type=int,
    default=100,
    help="player's default volume (default: 100)",
)
playerGroup.add_argument(
    "--default-crossfade",
    type=float,
    default=10.0,
    help="player's default crossfade seconds (default: 10.0)",
)
playerGroup.add_argument(
    "--default-gapless",
    type=bool,
    default=False,
    help="player's default gapless state (default: False)",
)
playerGroup.add_argument(
    "--default-autoplay",
    type=bool,
    default=True,
    help="player's default auto related play state (default: True)",
)
playerGroup.add_argument(
    "--bufferlimit",
    type=int,
    default=5,
    help="seconds of audio will be loaded in buffer (default: 5)",
)
playerGroup.add_argument(
    "--preload",
    type=int,
    default=10,
    help="seconds to load next song before this song ends (default: 10)",
)
playerGroup.add_argument(
    "--timeout",
    type=int,
    default=300,
    help="seconds to cleanup player when connection of discord terminated (default: 300)",
)

extExtractorParser = parser.add_argument_group("Extra Extractor Option")

extExtractorParser.add_argument(
    "--enabled-extractor",
    action="append",
    default=[],
    help="Extra extractors to enable (Support melon and spotify)",
)
extExtractorParser.add_argument(
    "--spotify-id",
    type=str,
    default=None,
    help="Spotify API id (default: None)",
)
extExtractorParser.add_argument(
    "--spotify-secret",
    type=str,
    default=None,
    help="Spotify API secret (default: None)",
)

logParser = parser.add_argument_group("Logging Option")

logParser.add_argument(
    "--verbose", "-v", action="store_true", help="Print various debugging information"
)

args = parser.parse_args()

if not args.config:
    verbose = args.verbose

    Config.HOST = args.host
    Config.PORT = args.port
    Config.PASSWORD = args.auth
    Config.HANDSHAKE_INTERVAL = args.ws_interval
    Config.HANDSHAKE_TIMEOUT = args.ws_timeout
    Config.IPBLOCKS = args.ip
    Config.EXCLUDEIPS = args.exclude_ip
    Config.DEFAULT_AUTOPLAY = args.default_autoplay
    Config.DEFAULT_VOLUME = round(args.default_volume / 100, 3)
    Config.DEFAULT_CROSSFADE = args.default_crossfade
    Config.DEFAULT_GAPLESS = args.default_gapless
    Config.BUFFERLIMIT = args.bufferlimit
    Config.VCTIMEOUT = args.timeout
    Config.ENABLED_EXT_EXTRACTOR = args.enabled_extractor
    Config.SPOTIFY_ID = args.spotify_id
    Config.SPOTIFY_SECRET = args.spotify_secret
else:
    verbose = args.config.pop("verbose", False)

    Config.from_dict(args.config)

if verbose:
    setLoggingLevel(logging.DEBUG)
else:
    setLoggingLevel(logging.INFO)

if __name__ == "__main__":
    try:
        import uvloop  # type: ignore
    except ModuleNotFoundError:
        pass
    else:
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

    loop = asyncio.get_event_loop()

    if log.level == logging.DEBUG:
        loop.set_debug(True)

    from .server import server

    loop.create_task(
        server.create_server(
            host=Config.HOST, port=Config.PORT, return_asyncio_server=True
        )
    )
    loop.run_forever()



import logging
import random
import re
from typing import Union

from youtube_related import RateLimited
from youtube_related import preventDuplication as relatedClient

from .config import Config
from .errors import NotPlaying
from .player import Player
from .source import AudioData, AudioSource
from .utils import EventDispatcher
from .voice_connector import VoiceConnector

log = logging.getLogger("discodo.VoiceClient")


class VoiceClient(VoiceConnector):
    YOUTUBE_VIDEO_REGEX = re.compile(
        r"^((?:https?:)?\/\/)?((?:www|m)\.)?((?:youtube\.com|youtu.be))(\/(?:[\w\-]+\?v=|embed\/|v\/)?)([\w\-]+)(\S+)?$"
    )

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.relatedClient = relatedClient()

        self.dispatcher = EventDispatcher()
        self.dispatcher.onAny(self.__dispatchToManager)
        self.dispatcher.on("REQUIRE_NEXT_SOURCE", self.__fetchAutoPlay)

        self.Queue = []
        self.player = None

        self._filter = {}
        self.paused = False

        self._autoplay = Config.DEFAULT_AUTOPLAY
        self._crossfade = Config.DEFAULT_CROSSFADE
        self._gapless = Config.DEFAULT_GAPLESS

        self._volume = Config.DEFAULT_VOLUME

        self.dispatcher.dispatch("VC_CREATED", id=self.id)

    def __del__(self) -> None:
        guild_id = int(self.guild_id) if self.guild_id else None

        log.info(f"destroying voice client of {guild_id}.")

        if self.manager.voiceClients.get(guild_id) == self:
            self.dispatcher.dispatch("VC_DESTROYED")
            del self.manager.voiceClients[guild_id]

        super().__del__()

        if self.player and self.player.is_alive():
            self.player.stop()

        for Item in self.Queue:
            if isinstance(Item, AudioSource):
                self.loop.call_soon_threadsafe(Item.cleanup)

    def __repr__(self) -> str:
        return f"<VoiceClient guild_id={self.guild_id} volume={self.volume} crossfade={self.crossfade} autoplay={self.autoplay} gapless={self.gapless}>"

    def __dispatchToManager(self, event, *args, **kwargs) -> None:
        self.manager.dispatcher.dispatch(self.guild_id, *args, event=event, **kwargs)

    async def __fetchAutoPlay(self, **kwargs):
        current = list(kwargs.values()).pop()

        for _ in range(5):
            if self.autoplay and not self.Queue:
                if self.YOUTUBE_VIDEO_REGEX.match(current.webpage_url):
                    address = Config.RoutePlanner.get() if Config.RoutePlanner else None
                    try:
                        Related = await self.relatedClient.async_get(
                            current.webpage_url, local_addr=address
                        )
                    except RateLimited:
                        Config.RoutePlanner.mark_failed_address(address)
                    else:
                        return await self.loadSource(Related["id"], related=True)

    def __spawnPlayer(self) -> None:
        if self.player and self.player.is_alive():
            return

        self.player = Player(self)
        self.player.crossfade = self._crossfade
        self.player.gapless = self._gapless

        self.player.start()

    async def createSocket(self, data: dict = None) -> None:
        """
        Create UDP socket with discord to send voice packet.

        :param dict data: A VOICE_SERVER_UPDATE payload, defaults to None
        :rtype: None
        """

        await super().createSocket(data)

        self.__spawnPlayer()

    @property
    def state(self) -> str:
        """
        The current state of playback of the voice client

        :getter: Returns VoiceClient state
        :rtype: str
        """

        if not self.Queue and not self.player.current:
            return "stopped"
        elif self.paused:
            return "paused"
        else:
            return "playing"

    @property
    def volume(self) -> float:
        """
        Current volume of the voice client (0.0~2.0)

        :getter: Returns VoiceClient volume
        :setter: Set VoiceClient volume
        :rtype: float
        """

        return self._volume

    @volume.setter
    def volume(self, value: float) -> None:
        self._volume = round(max(value, 0.0), 2)

    @property
    def crossfade(self) -> float:
        """
        Current crossfade value of the voice client

        :getter: Returns VoiceClient crossfade
        :setter: Set VoiceClient crossfade
        :rtype: float
        """

        return self._crossfade

    @crossfade.setter
    def crossfade(self, value: float) -> None:
        if not isinstance(value, float):
            return TypeError("`filter` property must be `float`.")

        self.player.crossfade = self._crossfade = round(max(value, 0.0), 1)

    @property
    def gapless(self) -> bool:
        """
        Current gapless value of the voice client

        :getter: Returns VoiceClient gapless
        :setter: Set VoiceClient gapless
        :rtype: bool
        """

        return self._gapless

    @gapless.setter
    def gapless(self, value: bool) -> None:
        if not isinstance(value, bool):
            return TypeError("`gapless` property must be `bool`.")

        self.player.gapless = self._gapless = self._gapless

    @property
    def autoplay(self) -> bool:
        """
        Current autoplay value of the voice client

        :getter: Returns VoiceClient autoplay
        :setter: Set VoiceClient autoplay
        :rtype: bool
        """

        return self._autoplay

    @autoplay.setter
    def autoplay(self, value: bool) -> None:
        if not isinstance(value, bool):
            return TypeError("`autoplay` property must be `bool`.")

        self._autoplay = value

    @property
    def filter(self) -> dict:
        """
        Current filter value of the voice client

        :getter: Returns VoiceClient filter
        :setter: Set VoiceClient filter
        :rtype: dict
        """

        return self._filter

    @filter.setter
    def filter(self, value: dict) -> None:
        if not isinstance(value, dict):
            return TypeError("`filter` property must be `dict`.")

        self._filter = value

    @property
    def current(self) -> Union[AudioSource, AudioData]:
        """
        Current source of the voice client

        :getter: Returns current source
        :rtype: :class:`AudioSource` or :class:`AudioData`
        """

        return self.player.current

    @property
    def next(self) -> Union[AudioSource, AudioData]:
        """
        Next source of the voice client

        :getter: Returns next source
        :rtype: :class:`AudioSource` or :class:`AudioData`
        """

        return self.player.next

    def putSource(self, Source: Union[list, AudioData, AudioSource]) -> int:
        """
        Put source to Queue.

        :param Source: list or :class:`AudioSource` or :class:`AudioData` to put on Queue
        :rtype: int
        """

        if not isinstance(Source, (list, AudioData, AudioSource)):
            raise TypeError("`Source` must be `list` or `AudioData` or `AudioSource`.")

        self.Queue += Source if isinstance(Source, list) else [Source]

        self.dispatcher.dispatch(
            "putSource", sources=(Source if isinstance(Source, list) else [Source])
        )

        return (
            self.Queue.index(Source)
            if not isinstance(Source, list)
            else [self.Queue.index(Item) for Item in Source]
        )

    async def getSource(self, Query: str) -> AudioData:
        """
        Get source from Query

        :param str Query: Query to search
        :rtype: :class:`AudioData`
        """

        return await AudioData.create(Query)

    async def loadSource(self, Query: str, **kwargs) -> AudioData:
        """
        Get source from Query and put it on Queue.

        :param str Query: Query to search
        :rtype: :class:`AudioData`
        """

        Data = await self.getSource(Query)

        self.dispatcher.dispatch(
            "loadSource", source=(Data if isinstance(Data, list) else Data), **kwargs
        )

        self.putSource(Data)

        return Data

    async def seek(self, offset: int) -> None:
        """
        Seek current playing source.

        :param float offset: offset to seek
        :rtype: None
        """

        return await self.player.seek(offset)

    def skip(self, offset: int = 1) -> None:
        """
        skip current playing source.

        :param float offset: offset to skip
        :rtype: None
        """

        if not self.player.current:
            raise NotPlaying

        if len(self.Queue) < (offset - 1):
            raise ValueError("`offset` is bigger than `Queue` size.")

        if offset > 1:
            del self.Queue[0 : (offset - 1)]

        self.player.current.stop()

    def pause(self) -> bool:
        """
        Pause playing

        :rtype: True
        """

        self.paused = True
        return True

    def resume(self) -> bool:
        """
        Resume playing

        :rtype: True
        """

        self.paused = False
        return False

    def shuffle(self) -> dict:
        """
        Shuffle Queue.

        :rtype: dict
        """

        if not self.Queue:
            raise ValueError("`Queue` is empty now.")

        random.shuffle(self.Queue)

        return self.Queue



import asyncio
import ipaddress
from typing import Union

import aiohttp

from ..config import Config
from .melon import get_query as melon_get_query
from .spotify import get_query as spotify_get_query
from .youtube_dl import extract as youtube_dl_extract


async def extract(
    query: Union[str, list],
    address: Union[ipaddress.IPv4Address, ipaddress.IPv6Address] = None,
    **kwargs
) -> dict:
    connector = aiohttp.TCPConnector(local_addr=(str(address), 0)) if address else None

    if "melon" in Config.ENABLED_EXT_EXTRACTOR:
        query = await melon_get_query(query, connector)
    if "spotify" in Config.ENABLED_EXT_EXTRACTOR:
        query = await spotify_get_query(query)

    if isinstance(query, list):
        Done, _ = await asyncio.wait(
            [youtube_dl_extract(Item, address=address, **kwargs) for Item in query]
        )

        Results = []
        for Future in Done:
            try:
                Results.append(Future.result())
            except:
                pass

        return Results

    return await youtube_dl_extract(query, address=address, **kwargs)



import asyncio
import uuid

from discodo.source import SubtitleSource

from ..errors import NotConnected, NotPlaying
from ..source import AudioData
from ..status import getStatus


def need_manager(func):
    def wrapper(self, *args, **kwargs):
        if not self.ClientManager:
            payload = {
                "op": func.__name__,
                "d": {"traceback": {"NOT_IDENTIFIED": "Identify first."}},
            }

            return self.sendJson(payload)
        return func(self, *args, **kwargs)

    return wrapper


class WebsocketPayloads:
    async def GET_STATUS(self, Data: None) -> None:
        payload = {"op": "STATUS", "d": getStatus()}

        await self.sendJson(payload)

    async def HEARTBEAT(self, Data: int) -> None:
        payload = {"op": "HEARTBEAT_ACK", "d": Data}

        await self.sendJson(payload)

    async def IDENTIFY(self, Data: dict) -> None:
        if self.ClientManager:
            payload = {
                "op": "IDENTIFY",
                "d": {
                    "traceback": {
                        "ALREADY_IDENTIFIED": "This connection already identified."
                    }
                },
            }
        else:
            await self.initialize_manager(Data["user_id"])
            payload = {"op": "IDENTIFIED", "d": "ClientManager initialized."}

        await self.sendJson(payload)

    @need_manager
    async def DISCORD_EVENT(self, Data: dict) -> None:
        self.ClientManager.discordDispatch(Data)

    @need_manager
    async def getSource(self, Data: dict) -> None:
        payload = {
            "op": "getSource",
            "d": {
                "source": await AudioData.create(Data["query"]),
            },
        }

        await self.sendJson(payload)

    @need_manager
    async def putSource(self, Data: dict) -> None:
        VoiceClient = self.ClientManager.getVC(Data["guild_id"])
        if not VoiceClient:
            raise NotConnected

        payload = {
            "op": "putSource",
            "d": {
                "guild_id": Data["guild_id"],
                "index": VoiceClient.putSource(Data["source"]),
            },
        }

        await self.sendJson(payload)

    @need_manager
    async def loadSource(self, Data: dict) -> None:
        VoiceClient = self.ClientManager.getVC(Data["guild_id"])
        if not VoiceClient:
            raise NotConnected

        await VoiceClient.loadSource(Data["query"])

    @need_manager
    async def setVolume(self, Data: dict) -> None:
        VoiceClient = self.ClientManager.getVC(Data["guild_id"])
        if not VoiceClient:
            raise NotConnected

        VoiceClient.volume = Data["volume"]

        payload = {
            "op": "setVolume",
            "d": {"guild_id": Data["guild_id"], "volume": VoiceClient.volume},
        }

        await self.sendJson(payload)

    @need_manager
    async def setCrossfade(self, Data: dict) -> None:
        VoiceClient = self.ClientManager.getVC(Data["guild_id"])
        if not VoiceClient:
            raise NotConnected

        VoiceClient.crossfade = Data["crossfade"]

        payload = {
            "op": "setCrossfade",
            "d": {"guild_id": Data["guild_id"], "crossfade": VoiceClient.crossfade},
        }

        await self.sendJson(payload)

    @need_manager
    async def setGapless(self, Data: dict) -> None:
        VoiceClient = self.ClientManager.getVC(Data["guild_id"])
        if not VoiceClient:
            raise NotConnected

        VoiceClient.gapless = Data["gapless"]

        payload = {
            "op": "setGapless",
            "d": {"guild_id": Data["guild_id"], "gapless": VoiceClient.gapless},
        }

        await self.sendJson(payload)

    @need_manager
    async def setAutoplay(self, Data: dict) -> None:
        VoiceClient = self.ClientManager.getVC(Data["guild_id"])
        if not VoiceClient:
            raise NotConnected

        VoiceClient.autoplay = Data["autoplay"]

        payload = {
            "op": "setAutoplay",
            "d": {"guild_id": Data["guild_id"], "autoplay": VoiceClient.autoplay},
        }

        await self.sendJson(payload)

    @need_manager
    async def setFilter(self, Data: dict) -> None:
        VoiceClient = self.ClientManager.getVC(Data["guild_id"])
        if not VoiceClient:
            raise NotConnected

        VoiceClient.filter = Data["filter"]

        payload = {
            "op": "setFilter",
            "d": {"guild_id": Data["guild_id"], "filter": VoiceClient.filter},
        }

        await self.sendJson(payload)

    @need_manager
    async def seek(self, Data: dict) -> None:
        VoiceClient = self.ClientManager.getVC(Data["guild_id"])
        if not VoiceClient:
            raise NotConnected

        await VoiceClient.seek(Data["offset"])

        payload = {
            "op": "seek",
            "d": {"guild_id": Data["guild_id"], "offset": Data["offset"]},
        }

        await self.sendJson(payload)

    @need_manager
    async def skip(self, Data: dict) -> None:
        VoiceClient = self.ClientManager.getVC(Data["guild_id"])
        if not VoiceClient:
            raise NotConnected

        VoiceClient.skip(Data["offset"])

        payload = {
            "op": "skip",
            "d": {"guild_id": Data["guild_id"], "remain": len(VoiceClient.Queue)},
        }

        await self.sendJson(payload)

    @need_manager
    async def pause(self, Data: dict) -> None:
        VoiceClient = self.ClientManager.getVC(Data["guild_id"])
        if not VoiceClient:
            raise NotConnected

        VoiceClient.pause()

        payload = {
            "op": "pause",
            "d": {"guild_id": Data["guild_id"], "state": VoiceClient.state},
        }

        await self.sendJson(payload)

    @need_manager
    async def resume(self, Data: dict) -> None:
        VoiceClient = self.ClientManager.getVC(Data["guild_id"])
        if not VoiceClient:
            raise NotConnected

        VoiceClient.resume()

        payload = {
            "op": "resume",
            "d": {"guild_id": Data["guild_id"], "state": VoiceClient.state},
        }

        await self.sendJson(payload)

    @need_manager
    async def shuffle(self, Data: dict) -> None:
        VoiceClient = self.ClientManager.getVC(Data["guild_id"])
        if not VoiceClient:
            raise NotConnected

        VoiceClient.shuffle()

        payload = {
            "op": "shuffle",
            "d": {"guild_id": Data["guild_id"], "entries": VoiceClient.Queue},
        }

        await self.sendJson(payload)

    @need_manager
    async def remove(self, Data: dict) -> None:
        VoiceClient = self.ClientManager.getVC(Data["guild_id"])
        if not VoiceClient:
            raise NotConnected

        removed = VoiceClient.Queue[Data["index"] - 1]
        del VoiceClient.Queue[Data["index"] - 1]

        payload = {
            "op": "shuffle",
            "d": {
                "guild_id": Data["guild_id"],
                "removed": removed,
                "entries": VoiceClient.Queue,
            },
        }

        await self.sendJson(payload)

    @need_manager
    async def getState(self, Data: dict) -> None:
        VoiceClient = self.ClientManager.getVC(Data["guild_id"])
        if not VoiceClient:
            raise NotConnected

        payload = {
            "op": "getState",
            "d": {
                "id": VoiceClient.id,
                "guild_id": VoiceClient.guild_id,
                "channel_id": VoiceClient.channel_id,
                "state": VoiceClient.state,
                "current": VoiceClient.current,
                "duration": VoiceClient.current.duration,
                "position": VoiceClient.current.position,
                "remain": VoiceClient.current.remain,
                "options": {
                    "autoplay": VoiceClient.autoplay,
                    "volume": VoiceClient.volume,
                    "crossfade": VoiceClient.crossfade,
                    "gapless": VoiceClient.gapless,
                    "filter": VoiceClient.filter,
                },
            },
        }

        return await self.sendJson(payload)

    @need_manager
    async def getQueue(self, Data: dict) -> None:
        VoiceClient = self.ClientManager.getVC(Data["guild_id"])
        if not VoiceClient:
            raise NotConnected

        payload = {
            "op": "getQueue",
            "d": {
                "guild_id": Data["guild_id"],
                "entries": VoiceClient.Queue,
            },
        }

        return await self.sendJson(payload)

    @need_manager
    async def VC_DESTROY(self, Data: dict) -> None:
        self.ClientManager.delVC(Data["guild_id"])

    @need_manager
    async def requestSubtitle(self, Data: dict) -> None:
        vc = self.ClientManager.getVC(Data["guild_id"])

        if not vc.current:
            raise NotPlaying

        if "url" in Data:
            url = Data["url"]
        else:
            if not Data["lang"] in vc.current.subtitles:
                payload = {
                    "op": "requestSubtitle",
                    "d": {
                        "guild_id": Data["guild_id"],
                        "NoSubtitle": f"There is no subtitle in {Data['lang']}",
                    },
                }
                return await self.sendJson(payload)

            url = vc.current.subtitles[Data["lang"]]

        current = vc.current
        _identify_token = str(uuid.uuid4())

        if url.endswith(".smi"):
            Subtitle = await SubtitleSource.smi.load(url)
        else:
            Subtitle = await SubtitleSource.srv1.load(url)

        payload = {
            "op": "requestSubtitle",
            "d": {
                "guild_id": Data["guild_id"],
                "identify": _identify_token,
                "url": url,
            },
        }
        await self.sendJson(payload)

        Previous = Now = Next = ""
        Elements = list(Subtitle.TextElements.values())
        while not Subtitle.is_done and not current.stopped:
            Element = Subtitle.seek(current.position)

            if Element and Element["markdown"] and Element["markdown"] != Now:
                Previous = Now
                Now = Element["markdown"]

                NextElements = [
                    NextElement["markdown"]
                    for NextElement in Elements[Elements.index(Element) + 1 :]
                    if NextElement["markdown"] != Now
                ]

                Next = NextElements[0] if NextElements else None

                payload = {
                    "op": "Subtitle",
                    "d": {
                        "guild_id": Data["guild_id"],
                        "identify": _identify_token,
                        "previous": Previous,
                        "current": Now,
                        "next": Next,
                    },
                }
                await self.sendJson(payload)

            await asyncio.sleep(0.1)

        payload = {
            "op": "subtitleDone",
            "d": {
                "guild_id": Data["guild_id"],
                "identify": _identify_token,
                "language": Data["language"],
            },
        }
        return await self.sendJson(payload)



import logging

from .utils import EventDispatcher
from .voice_client import VoiceClient

log = logging.getLogger("discodo.event")


class DiscordEvent:
    dispatcher = EventDispatcher()

    def __init__(self, manager) -> None:
        self.manager = manager

        self.dispatcher.on("READY", self.READY)
        self.dispatcher.on("RESUME", self.RESUME)
        self.dispatcher.on("VOICE_STATE_UPDATE", self.VOICE_STATE_UPDATE)
        self.dispatcher.on("VOICE_SERVER_UPDATE", self.VOICE_SERVER_UPDATE)

    def dispatch(self, event: str, *args, **kwargs) -> None:
        return self.dispatcher.dispatch(event, *args, **kwargs)

    async def READY(self, data: dict) -> None:
        log.info(
            f'ready event dispatched. set session id {data["session_id"]} and user id {data["user"]["id"]}'
        )
        self.manager.session_id = data["session_id"]
        self.manager.id = int(data["user"]["id"])

    async def RESUME(self, data: dict) -> None:
        log.info(
            f'resume event dispatched. set session id {data["session_id"]} and user id {data["user"]["id"]}'
        )
        self.manager.session_id = data["session_id"]
        self.manager.id = int(data["user"]["id"])

    async def VOICE_STATE_UPDATE(self, data: dict) -> None:
        if not self.manager.id or data["user_id"] != str(self.manager.id):
            return

        log.info(f'recieve self voice update. set session id {data["session_id"]}')
        self.manager.session_id = data["session_id"]

        if self.manager.getVC(data["guild_id"], safe=True):
            self.manager.getVC(data["guild_id"]).channel_id = (
                int(data["channel_id"]) if data.get("channel_id") else None
            )
        elif data.get("channel_id"):
            log.info(f'Voice Client of {data["guild_id"]} not found. create new one.')

            vc = self.manager.voiceClients[int(data["guild_id"])] = VoiceClient(
                self.manager
            )
            vc.guild_id = int(data["guild_id"])
            vc.channel_id = int(data["channel_id"])

    async def VOICE_SERVER_UPDATE(self, data: dict) -> None:
        if self.manager.getVC(data["guild_id"], safe=True):
            log.info(f"Voice server update recieved. connect to new endpoint")

            await self.manager.getVC(data["guild_id"]).createSocket(data)
        else:
            log.info(f'Voice Client of {data["guild_id"]} not found. create new one.')

            self.manager.voiceClients[int(data["guild_id"])] = VoiceClient(
                self.manager, data
            )



import asyncio
import logging
import re
import socket
import struct
import threading
import uuid

from websockets.exceptions import ConnectionClosed

from .config import Config
from .gateway import VoiceSocket
from .natives import Encrypter, opus

log = logging.getLogger("discodo.VoiceConnector")


class VoiceConnector:
    def __init__(self, manager, data: dict = {}) -> None:
        self.loop = manager.loop

        self.ws = self.socket = None

        self.id = str(uuid.uuid4())
        self.manager = manager

        self.data = data
        self.guild_id = data.get("guild_id")
        self.channel_id = None

        self._connected = asyncio.Event()
        self._connectedThread = threading.Event()

        self._polling = None
        self._sequence = self._timestamp = 0
        self.encoder = opus.Encoder()

        if data:
            self.loop.create_task(self.createSocket())

    def __del__(self) -> None:
        if self.socket:
            try:
                self.socket.close()
            except:
                pass

        if self.ws:
            self.loop.create_task(self.ws.close(4000))

        if self._polling:
            self._polling.cancel()

    def __repr__(self) -> str:
        return f"<VoiceConnector guild_id={self.guild_id} connected={self._connected.is_set()} sequence={self.sequence} timestamp={self.timestamp}>"

    @property
    def user_id(self) -> int:
        """
        User id which is connected.

        :getter: Returns User id
        :rtype: int
        """

        return self.manager.id

    @property
    def session_id(self) -> str:
        """
        Session id which is connected.

        :getter: Returns Session id
        :rtype: str
        """

        return self.manager.session_id

    @property
    def sequence(self) -> int:
        """
        Sequence value which is used to send voice packet.

        :getter: Returns sequence
        :setter: Sets sequence value (max 65535)
        :rtype: int
        """

        return self._sequence

    @sequence.setter
    def sequence(self, value: int) -> None:
        if self._sequence + value > 65535:
            self._sequence = 0
        else:
            self._sequence = value

    @property
    def timestamp(self) -> int:
        """
        Timestamp value which is used to send voice packet.

        :getter: Returns timestamp
        :setter: Sets timestamp value (max 4294967295)
        :rtype: int
        """

        return self._timestamp

    @timestamp.setter
    def timestamp(self, value: int) -> None:
        if self._timestamp + value > 4294967295:
            self._timestamp = 0
        else:
            self._timestamp = value

    async def createSocket(self, data: dict = None) -> None:
        """
        Create UDP socket with discord to send voice packet.

        :param dict data: A VOICE_SERVER_UPDATE payload, defaults to None
        :rtype: None
        """

        self._connected.clear()
        self._connectedThread.clear()

        if data:
            self.data = data

        self.guild_id = self.data.get("guild_id")

        self.token = self.data.get("token")
        self.endpoint = re.sub(r"(:[0-9]+)", "", self.data.get("endpoint"))
        log.info(f"voice endpoint {self.endpoint} detected.")

        if self.socket:
            try:
                self.socket.close()
            except:
                pass

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.setblocking(False)

        if self.ws:
            await self.ws.close(4000)

        if hasattr(self, "secretKey"):
            del self.secretKey

        self.ws = await VoiceSocket.connect(self)
        while not hasattr(self, "secretKey"):
            await self.ws.poll()

        if not self._polling or self._polling.done():
            self._polling = self.loop.create_task(self.pollingWS())

        self._connected.set()
        self._connectedThread.set()

    async def pollingWS(self) -> None:
        while True:
            try:
                await self.ws.poll()
            except (asyncio.TimeoutError, ConnectionClosed) as e:
                self._connected.clear()
                self._connectedThread.clear()

                reason = (
                    f"with {e.code}"
                    if isinstance(e, ConnectionClosed)
                    else "because timed out."
                )

                log.info(
                    f"voice connection of {self.guild_id} destroyed {reason}. wait for events."
                )

                try:
                    await asyncio.wait_for(
                        self._connected.wait(), timeout=Config.VCTIMEOUT
                    )
                except asyncio.TimeoutError:
                    return self.__del__()

    def makePacket(self, data: bytes) -> bytes:
        """
        Converts and encrypts data to format.

        :param bytes data: A packet to be converted
        :rtype: bytes
        """

        header = bytearray(12)
        header[0], header[1] = 0x80, 0x78

        struct.pack_into(">H", header, 2, self.sequence)
        struct.pack_into(">I", header, 4, self.timestamp)
        struct.pack_into(">I", header, 8, self.ssrc)

        return getattr(Encrypter, self.encryptMode)(self.secretKey, header, data)

    def send(self, data: bytes, encode: bool = True) -> None:
        """
        Send packet to discord.

        :param bytes data: A packet to be sent
        :param bool encode: Whether data is encoded
        :rtype: None
        """

        self.sequence += 1
        if encode:
            data = self.encoder.encode(data)

        Packet = self.makePacket(data)

        self.socket.sendto(Packet, (self.endpointIP, self.endpointPORT))
        self.timestamp += Config.SAMPLES_PER_FRAME



from typing import Coroutine

from sanic import Sanic, response
from sanic.exceptions import abort

from .. import __version__
from ..config import Config
from ..status import getStatus
from .planner import app as PlannerBlueprint
from .restful import app as RestfulBlueprint
from .websocket import app as WebsocketBlueprint

app = Sanic(__name__)

app.register_blueprint(WebsocketBlueprint)
app.register_blueprint(PlannerBlueprint)
app.register_blueprint(RestfulBlueprint)


def authorized(func: Coroutine) -> Coroutine:
    def wrapper(request, *args, **kwargs) -> Coroutine:
        if request.headers.get("Authorization") != Config.PASSWORD:
            abort(403, "Password mismatch.")

        return func(request, *args, **kwargs)

    return wrapper


@app.route("/")
async def index(request) -> response:
    return response.html(f"<h1>Discodo</h1> <h3>{__version__}")


@app.get("/status")
async def status(request) -> response:
    return response.json(getStatus())



import nacl.secret


class Encrypter:
    available = ["xsalsa20_poly1305", "xsalsa20_poly1305_suffix"]

    def xsalsa20_poly1305(key: list, header: bytearray, data: bytes) -> bytes:
        Box = nacl.secret.SecretBox(bytes(key))
        Nonce = bytearray(24)
        Nonce[:12] = header

        return header + Box.encrypt(bytes(data), bytes(Nonce)).ciphertext

    def xsalsa20_poly1305_suffix(key: list, header: bytearray, data: bytes) -> bytes:
        Box = nacl.secret.SecretBox(bytes(key))
        Nonce = nacl.utils.random(nacl.secret.SecretBox.NONCE_SIZE)

        return header + Box.encrypt(bytes(data), Nonce).ciphertext + Nonce



__version__ = "2.1.1"

from .client import *
from .config import Config
from .errors import *
from .event import DiscordEvent
from .extractor import extract
from .manager import ClientManager
from .natives import *
from .player import Player
from .source import *
from .status import getStatus
from .utils import *
from .voice_client import VoiceClient



import asyncio
import logging
from typing import Coroutine

import websockets

from ..errors import NodeNotConnected, VoiceClientNotFound
from ..utils import EventDispatcher
from .gateway import NodeConnection
from .voice_client import VoiceClient

log = logging.getLogger("discodo.client")


class Node:
    def __init__(
        self,
        host: str,
        port: int,
        user_id: int = None,
        password: str = "hellodiscodo",
        region: str = None,
    ) -> None:
        self.ws = None
        self.dispatcher = EventDispatcher()
        self.dispatcher.onAny(self.onAnyEvent)

        self.loop = asyncio.get_event_loop()

        self.host = host
        self.port = port
        self.password = password

        self.user_id = user_id
        self.region = region

        self._polling = None
        self.voiceClients = {}
        self.connected = asyncio.Event()

    @property
    def URL(self) -> str:
        return f"ws://{self.host}:{self.port}/ws"

    async def connect(self) -> None:
        if self.connected.is_set():
            raise ValueError("Node already connected")

        if self.ws and not self.ws.closed:
            await self.ws.close()

        self.ws = await NodeConnection.connect(self)
        self.voiceClients = {}
        self.connected.set()

        if not self._polling or self._polling.done():
            self._polling = self.loop.create_task(self.pollingWS())

        if self.user_id:
            await self.ws.send("IDENTIFY", {"user_id": self.user_id})

    @property
    def is_connected(self) -> bool:
        return self.connected.is_set() and self.ws and self.ws.is_connected

    async def destroy(self) -> None:
        if self._polling and not self._polling.done():
            self._polling.cancel()

        if self.ws and not self.ws.closed:
            await self.ws.close()

        self.connected.clear()
        self.voiceClients = {}
        self.ws = None

    async def pollingWS(self) -> None:
        while True:
            try:
                Operation, Data = await self.ws.poll()
            except (asyncio.TimeoutError, websockets.ConnectionClosedError):
                self.connected.clear()

                if self.ws:
                    await self.ws.close()

                self.ws = None
                self.voiceClients = {}
                return

            log.debug(f"event {Operation} dispatched from websocket with {Data}")

            self.dispatcher.dispatch(Operation, Data)

    def send(self, *args, **kwargs) -> Coroutine:
        if not self.ws:
            raise NodeNotConnected

        return self.ws.send(*args, **kwargs)

    async def onAnyEvent(self, Operation, Data):
        if Operation == "RESUMED":
            for voice_client in self.voiceClients:
                voice_client.__del__()

            for guild_id, _ in Data["channels"].items():
                self.voiceClients[int(guild_id)] = VoiceClient(self, guild_id)

        if Operation == "VC_CREATED":
            guild_id = int(Data["guild_id"])
            self.voiceClients[guild_id] = VoiceClient(self, guild_id)

        if Data and isinstance(Data, dict) and "guild_id" in Data:
            vc = self.getVC(Data["guild_id"])
            if vc:
                vc.dispatcher.dispatch(Operation, Data)

        if Operation == "VC_DESTROYED":
            guild_id = int(Data["guild_id"])
            if guild_id in self.voiceClients:
                self.voiceClients[guild_id].__del__()

    def getVC(self, guildID: int, safe: bool = False) -> VoiceClient:
        if not int(guildID) in self.voiceClients and not safe:
            raise VoiceClientNotFound

        return self.voiceClients.get(int(guildID))

    async def discordDispatch(self, payload: dict) -> None:
        if not payload["t"] in [
            "READY",
            "RESUME",
            "VOICE_STATE_UPDATE",
            "VOICE_SERVER_UPDATE",
        ]:
            return

        return await self.send("DISCORD_EVENT", payload)

    async def getStat(self) -> dict:
        await self.send("GET_STAT")

        return await self.dispatcher.wait_for("STAT", timeout=10.0)



import ipaddress
from typing import Coroutine

from sanic import Blueprint, response
from sanic.exceptions import abort

from ..config import Config

app = Blueprint(__name__)


def authorized(func: Coroutine) -> Coroutine:
    def wrapper(request, *args, **kwargs) -> Coroutine:
        if request.headers.get("Authorization") != Config.PASSWORD:
            abort(403, "Password mismatch.")

        return func(request, *args, **kwargs)

    return wrapper


@app.get("/planner")
@authorized
async def plannerStatus(request) -> response.json:
    if not Config.RoutePlanner:
        abort(404, "RoutePlanner is not enabled.")

    return response.json(
        {
            "ipBlocks": [
                {
                    "version": ipBlock.version,
                    "broadcast_address": ipBlock.broadcast_address,
                    "size": ipBlock.num_addresses,
                }
                for ipBlock in Config.RoutePlanner.ipBlocks
            ],
            "failedAddresses": [
                dict(data, address=address)
                for address, data in Config.RoutePlanner.failedAddress.items()
            ],
        }
    )


@app.post("/planner/unmark")
@authorized
async def plannerUnmark(request) -> response.json:
    if not Config.RoutePlanner:
        abort(404, "RoutePlanner is not enabled.")

    address = "".join(request.args.get("address", [])).strip()
    if not address:
        abort(400, "Missing parameter address.")

    Config.RoutePlanner.unmark_failed_address(ipaddress.ip_address(address))

    return response.json({"status": 200})


@app.post("/planner/unmark/all")
@authorized
async def plannerUnmark(request) -> response.json:
    if not Config.RoutePlanner:
        abort(404, "RoutePlanner is not enabled.")

    Config.RoutePlanner.failedAddress.clear()

    return response.json({"status": 200})



import asyncio
import audioop
import functools
import ipaddress
import threading
import traceback
import urllib.parse
from typing import Coroutine, Union

import av

from ..config import Config
from ..natives import AudioFifo, AudioFilter
from ..utils.threadLock import withLock

AVOption = {
    "err_detect": "ignore_err",
    "reconnect": "1",
    "reconnect_streamed": "1",
    "reconnect_delay_max": "5",
}


class PyAVSource:
    def __init__(self, Source: str) -> None:
        self.loop = asyncio.get_event_loop()

        self.Source = Source
        self.AVOption = AVOption
        self.Container: av.StreamContainer = None
        self.selectAudioStream = self.FrameGenerator = None

        self._end = threading.Event()
        self._haveToReloadResampler = threading.Event()
        self._waitforread = threading.Lock()
        self._loading = threading.Lock()
        self._seeking = threading.Lock()
        self.BufferLoader = None

        self.AudioFifo = AudioFifo()
        self._duration = None
        self._position = 0.0
        self._volume = 1.0
        self._filter = {}
        self.stopped = False

    def __del__(self):
        self.cleanup()

    @property
    def volume(self) -> float:
        return self._volume

    @volume.setter
    def volume(self, value: float) -> None:
        self._volume = max(value, 0.0)

    @property
    def filter(self) -> dict:
        return self._filter

    @filter.setter
    def filter(self, value: dict) -> None:
        self._filter = value

    @property
    def duration(self) -> float:
        return self._duration

    @property
    def position(self) -> float:
        return round(
            self._position
            - (
                self.AudioFifo.samples
                / 960
                / 50
                * (self.filter["atempo"] if "atempo" in self.filter else 1.0)
            ),
            2,
        )

    def read(self) -> bytes:
        if not self.BufferLoader:
            self.start()

        if not self.AudioFifo:
            return

        Data = self.AudioFifo.read()
        if not Data and self._loading.locked():
            while self._loading.locked():
                if not self._waitforread.locked():
                    self._waitforread.acquire()
                self._waitforread.acquire()

                Data = self.AudioFifo.read()
                if Data:
                    break

        if Data and self.volume != 1.0:
            Data = audioop.mul(Data, 2, min(self._volume, 2.0))

        return Data

    def _seek(self, offset: float, *args, **kwargs) -> None:
        with withLock(self._seeking):
            if not self.Container:
                if not self._loading.locked():
                    self.Container = av.open(
                        self.Source.Source, options=self.Source.AVOption
                    )
                else:
                    while not self.Container:
                        pass

            kwargs["any_frame"] = True

            self.Container.seek(round(max(offset, 1) * 1000000), *args, **kwargs)
            self.reload()

    def seek(self, offset: float, *args, **kwargs) -> Coroutine:
        return self.loop.run_in_executor(
            None, functools.partial(self._seek, offset, *args, **kwargs)
        )

    def reload(self) -> None:
        self._haveToReloadResampler.set()

        if not self._loading.locked():
            if self._end.is_set():
                self._end.clear()
            self.start()

    def start(self) -> None:
        self.BufferLoader = Loader(self)
        self.BufferLoader.start()

    def stop(self) -> bool:
        self.stopped = True
        return self.stopped

    def is_opus(self) -> bool:
        return False

    def cleanup(self) -> None:
        self._end.set()
        if self.AudioFifo and not self.AudioFifo.haveToFillBuffer.is_set():
            self.AudioFifo.haveToFillBuffer.clear()
        self.AudioFifo = None


class Loader(threading.Thread):
    def __init__(self, AudioSource: PyAVSource) -> None:
        threading.Thread.__init__(self)
        self.daemon = True

        self.Source = AudioSource

        self.Resampler = None
        self.Filter = {}
        self.FilterGraph = None

    def _do_run(self) -> None:
        with withLock(self.Source._loading):
            if not self.Source.Container:
                self.Source.Container = av.open(
                    self.Source.Source, options=self.Source.AVOption
                )
            self.Source._duration = round(self.Source.Container.duration / 1000000, 2)

            self.Source.selectAudioStream = self.Source.Container.streams.audio[0]
            self.Source.FrameGenerator = self.Source.Container.decode(
                self.Source.selectAudioStream
            )

            while not self.Source._end.is_set():
                if self.Source.filter != self.Filter:
                    self.Filter = self.Source.filter

                    if self.Source.filter:
                        self.FilterGraph = AudioFilter()
                        self.FilterGraph.selectAudioStream = (
                            self.Source.selectAudioStream
                        )
                        self.FilterGraph.setFilters(self.Filter)
                    else:
                        self.FilterGraph = None

                if not self.Resampler or self.Source._haveToReloadResampler.is_set():
                    self.Resampler = av.AudioResampler(
                        format=av.AudioFormat("s16").packed, layout="stereo", rate=48000
                    )
                    self.Source._haveToReloadResampler.clear()

                _seek_locked = False
                if self.Source._seeking.locked():
                    self.Source._seeking.acquire()
                    _seek_locked = True

                Frame = next(self.Source.FrameGenerator, None)

                if _seek_locked:
                    self.Source._seeking.release()
                    self.Source.AudioFifo.reset()

                if not Frame:
                    self.Source.stop()
                    break

                _current_position = float(Frame.pts * Frame.time_base)

                if self.FilterGraph:
                    self.FilterGraph.push(Frame)
                    Frame = self.FilterGraph.pull()

                    if not Frame:
                        continue

                Frame.pts = None
                try:
                    Frame = self.Resampler.resample(Frame)
                except ValueError:
                    self.Source._haveToReloadResampler.set()
                    continue

                if self.Source.AudioFifo:
                    if not self.Source.AudioFifo.haveToFillBuffer.is_set():
                        self.Source.AudioFifo.haveToFillBuffer.wait()

                    self.Source.AudioFifo.write(Frame)

                self.Source._position = _current_position

                if self.Source._waitforread.locked():
                    self.Source._waitforread.release()

    def run(self) -> None:
        try:
            self._do_run()
        except:
            traceback.print_exc()
        finally:
            if self.Source.Container:
                self.Source.Container.close()
                self.Source.Container = None

            self.Source.stop()



import asyncio
import logging
from itertools import chain
from typing import Union

import discord

from ..errors import NodeNotConnected, VoiceClientNotFound
from ..utils import EventDispatcher
from .node import Node as OriginNode
from .voice_client import VoiceClient

log = logging.getLogger("discodo.client")


class NodeClient(OriginNode):
    def __init__(self, DPYClient, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.DPYClient = DPYClient

    async def destroy(self, *args, **kwargs) -> None:
        log.infmo(f"destroying Node {self.URL}")
        await super().destroy(*args, **kwargs)

        if self in self.DPYClient.Nodes:
            self.DPYClient.Nodes.remove(self)


class DPYClient:
    def __init__(self, client) -> None:
        self.client = client
        self.loop = client.loop or asyncio.get_event_loop()

        self.dispatcher = EventDispatcher()
        self.event = self.dispatcher.event

        self.Nodes = []
        self.__register_event()

    def __register_event(self):
        if hasattr(self.client, "on_socket_response"):
            originFunc = self.client.on_socket_response
        else:
            originFunc = None

        @self.client.event
        async def on_socket_response(*args, **kwargs):
            self.loop.create_task(self.discord_socket_response(*args, **kwargs))

            if originFunc:
                return await originFunc()

    async def discord_socket_response(self, payload: dict) -> None:
        if payload["t"] == "VOICE_SERVER_UPDATE":
            VC = self.getVC(payload["d"]["guild_id"], safe=True)
            SelectNodes = [VC.Node] if VC else [self.getBestNode()]
        else:
            SelectNodes = self.Nodes

        if SelectNodes:
            await asyncio.wait(
                [Node.discordDispatch(payload) for Node in SelectNodes],
                return_when="ALL_COMPLETED",
            )

    def register_node(self, *args, **kwargs) -> None:
        return self.loop.create_task(self._register_event(*args, **kwargs))

    async def _register_event(self, *args, **kwargs) -> None:
        await self.client.wait_until_ready()
        kwargs["user_id"] = self.client.user.id

        Node = NodeClient(self, *args, **kwargs)
        await Node.connect()

        log.info(f"registering Node {Node.host}:{Node.port}")

        self.Nodes.append(Node)

        Node.dispatcher.on("RESUMED", self._resumed)
        Node.dispatcher.on("VC_DESTROYED", self._vc_destroyed)
        Node.dispatcher.onAny(self._node_event)

        return self

    async def _resumed(self, Data: dict) -> None:
        for guild_id, channel_id in Data["channels"].items():
            guild = self.client.get_guild(int(guild_id))
            if channel_id:
                channel = guild.get_channel(channel_id)
                self.loop.create_task(self.connect(channel))
            else:
                self.loop.create_task(self.disconnect(guild))

    async def _vc_destroyed(self, Data: dict) -> None:
        guild = self.client.get_guild(int(Data["guild_id"]))
        ws = self.__get_websocket(guild.shard_id)

        await ws.voice_state(guild.id, None)

    async def _node_event(self, Event: str, Data: dict) -> None:
        if not isinstance(Data, dict) or not "guild_id" in Data:
            return

        guild = self.client.get_guild(int(Data["guild_id"]))
        vc = self.getVC(guild)

        self.dispatcher.dispatch(Event, vc, Data)

    def getBestNode(self):
        SortedWithPerformance = sorted(
            [Node for Node in self.Nodes if Node.is_connected],
            key=lambda Node: len(Node.voiceClients),
        )

        return SortedWithPerformance[0] if SortedWithPerformance else None

    @property
    def voiceClients(self):
        return {
            ID: Value
            for ID, Value in list(
                chain.from_iterable(
                    [
                        Node.voiceClients.items()
                        for Node in self.Nodes
                        if Node.is_connected
                    ]
                )
            )
        }

    def getVC(
        self, guild: Union[discord.Guild, int], safe: bool = False
    ) -> VoiceClient:
        if isinstance(guild, discord.Guild):
            guild = guild.id

        if not int(guild) in self.voiceClients and not safe:
            raise VoiceClientNotFound

        return self.voiceClients.get(int(guild))

    def __get_websocket(self, id: int):
        if isinstance(self.client, discord.AutoShardedClient):
            return self.client.shards[id].ws
        elif not self.client.shard_id or self.client.shard_id == id:
            return self.client.ws

    async def connect(self, channel: discord.VoiceChannel) -> None:
        log.info(f"connecting to {channel.id} of {channel.guild.id}")
        if not hasattr(channel, "guild"):
            raise ValueError

        if not self.getBestNode():
            raise NodeNotConnected

        ws = self.__get_websocket(channel.guild.shard_id)

        await ws.voice_state(channel.guild.id, channel.id)

    async def disconnect(self, guild: discord.Guild) -> None:
        log.info(f"disconnecting voice of {guild.id} without destroying")
        ws = self.__get_websocket(guild.shard_id)

        await ws.voice_state(guild.id, None)

    async def destroy(self, guild: discord.Guild) -> None:
        log.info(f"destroying voice client of {guild.id}")
        if not guild.id in self.voiceClients:
            raise VoiceClientNotFound

        vc = self.getVC(guild.id)
        ws = self.__get_websocket(guild.shard_id)

        await ws.voice_state(guild.id, None)
        await vc.destroy()



"""
Thx for Sannoob
"""

import collections
import ipaddress
import random
import time
from typing import Union


class RoutePlanner:
    def __init__(self, ipBlocks: list, excludeIps: list = []) -> None:
        self.failedAddress = {}
        self.usedCount = collections.defaultdict(int)

        self.ipBlocks = [ipaddress.ip_network(ipBlock) for ipBlock in ipBlocks]
        self.excludeIps = [ipaddress.ip_address(excludeIp) for excludeIp in excludeIps]

    def mark_failed_address(
        self,
        address: Union[ipaddress.IPv4Address, ipaddress.IPv6Address],
        status: int = 429,
    ) -> None:
        self.failedAddress[address] = {"status": status, "failed_at": time.time()}

    def unmark_failed_address(
        self, address: Union[ipaddress.IPv4Address, ipaddress.IPv6Address]
    ) -> None:
        del self.failedAddress[address]

    def __get_random(
        self, ipBlock: Union[ipaddress.IPv4Network, ipaddress.IPv6Network]
    ) -> str:
        cidr = ipBlock.prefixlen

        host_bits = (128 if ipBlock.version == 6 else 32) - cidr

        start = (int(ipBlock.broadcast_address) >> host_bits) << host_bits
        end = start | ((1 << host_bits) - 1)

        return ipaddress.ip_address(random.randrange(start, end))

    def get(self) -> Union[ipaddress.IPv4Address, ipaddress.IPv6Address]:
        sortedIpBlocks = sorted(
            self.ipBlocks, key=lambda ipBlock: self.usedCount.get(ipBlock, 0)
        )
        randomResult = self.__get_random(sortedIpBlocks[0])
        self.usedCount[sortedIpBlocks[0]] += 1

        if len(self.failedAddress) >= sum(
            [ipBlock.num_addresses for ipBlock in self.ipBlocks]
        ):
            raise ValueError("No Ips available")

        if randomResult in self.excludeIps or randomResult in self.failedAddress:
            return self.get()

        return randomResult



import asyncio
import json
import logging
from typing import Coroutine

from sanic import Blueprint
from sanic.websocket import ConnectionClosed

from ..config import Config
from ..manager import ClientManager
from .payloads import WebsocketPayloads

log = logging.getLogger("discodo.server")

app = Blueprint(__name__)


@app.websocket("/ws")
async def socket_feed(request, ws) -> None:
    handler = WebsocketHandler(request, ws)
    await handler.join()


class Encoder(json.JSONEncoder):
    def default(self, o):
        return o.__dict__()


class ModifyClientManager(ClientManager):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._binded = asyncio.Event()


class WebsocketHandler:
    def __init__(self, request, ws) -> None:
        self.loop = asyncio.get_event_loop()

        self.request = request
        self.ws = ws
        self.app = request.app

        if not hasattr(self.app, "ClientManagers"):
            self.app.ClientManagers = {}

        self.ClientManager = None

        self._running = asyncio.Event()
        self.loop.create_task(self._handle())

    def __del__(self) -> None:
        if self.ClientManager:
            self.loop.create_task(self.wait_for_bind())

    async def wait_for_bind(self) -> None:
        self.ClientManager._binded.clear()

        try:
            await asyncio.wait_for(
                self.ClientManager._binded.wait(), timeout=Config.VCTIMEOUT
            )
        except asyncio.TimeoutError:
            self.ClientManager.__del__()
            del self.app.ClientManagers[int(self.ClientManager.id)]
            self.ClientManager = None

    def join(self) -> Coroutine:
        return self._running.wait()

    async def _handle(self) -> None:
        log.info(f"new websocket connection created from {self.request.ip}.")

        if self.request.headers.get("Authorization") != Config.PASSWORD:
            log.warning(
                f"websocket connection from {self.request.ip} forbidden: password mismatch."
            )
            return await self.forbidden("Password mismatch.")

        await self.hello()

        while True:
            try:
                RAWDATA = await asyncio.wait_for(
                    self.ws.recv(), timeout=Config.HANDSHAKE_TIMEOUT
                )
            except (asyncio.TimeoutError, ConnectionClosed) as exception:
                if isinstance(exception, asyncio.TimeoutError):
                    log.info("websocket connection closing because of timeout.")
                elif isinstance(exception, ConnectionClosed):
                    log.info(
                        f"websocket connection disconnected. code {exception.code}"
                    )

                return

            try:
                _Data = json.loads(RAWDATA)
                Operation, Data = _Data.get("op"), _Data.get("d")
            except:
                continue

            if Operation and hasattr(WebsocketPayloads, Operation):
                log.debug(f"{Operation} dispatched with {Data}")

                Func = getattr(WebsocketPayloads, Operation)
                self.loop.create_task(self.__run_event(Func, Operation, Data))

    async def __run_event(self, Func: Coroutine, Operation: str, Data) -> None:
        try:
            await Func(self, Data)
        except Exception as e:
            payload = {
                "op": Operation,
                "d": {"traceback": {type(e).__name__: str(e)}},
            }

            if "guild_id" in Data:
                payload["d"]["guild_id"] = Data["guild_id"]

            await self.sendJson(payload)

    async def sendJson(self, Data: dict) -> None:
        log.debug(f"send {Data} to websocket connection of {self.request.ip}.")
        await self.ws.send(json.dumps(Data, cls=Encoder))

    async def initialize_manager(self, user_id: int) -> None:
        if int(user_id) in self.app.ClientManagers:
            self.ClientManager = self.app.ClientManagers[int(user_id)]
            self.ClientManager._binded.set()

            self.loop.create_task(self.resumed())
            log.debug(f"ClientManager of {user_id} resumed.")
        else:
            self.ClientManager = ModifyClientManager(user_id=user_id)
            self.app.ClientManagers[int(user_id)] = self.ClientManager

            log.debug(f"ClientManager of {user_id} intalized.")

        self.ClientManager.dispatcher.onAny(self.manager_event)

    async def manager_event(self, guild_id: int, **kwargs) -> None:
        kwargs = {key: value for key, value in kwargs.items()}
        Event = kwargs.pop("event")

        payload = {"op": Event, "d": {"guild_id": guild_id}}
        payload["d"] = dict(payload["d"], **kwargs)

        await self.sendJson(payload)

    async def hello(self) -> None:
        payload = {
            "op": "HELLO",
            "d": {"heartbeat_interval": Config.HANDSHAKE_INTERVAL},
        }

        await self.sendJson(payload)

    async def forbidden(self, message: str) -> None:
        payload = {"op": "FORBIDDEN", "d": message}

        await self.sendJson(payload)

    async def resumed(self) -> None:
        payload = {
            "op": "RESUMED",
            "d": {
                "channels": {
                    guild_id: voiceClient.channel_id
                    for guild_id, voiceClient in self.ClientManager.voiceClients.items()
                }
            },
        }

        await self.sendJson(payload)



import asyncio
import audioop
import threading
import time
import traceback
from typing import Callable, Union

from .config import Config
from .errors import NotPlaying
from .source import AudioData, AudioSource


class Player(threading.Thread):
    def __init__(self, client) -> None:
        threading.Thread.__init__(self)
        self.daemon = True

        self.client = client

        self._volume = 1.0
        self._crossfade = 5.0
        self._gapless = False
        self._end = threading.Event()

        self._getSourceTask = None
        self._current = self._next = None

        self._crossfadeLoop = 0
        self._request_dispatched = False

    def __del__(self) -> None:
        for Source in self.client.Queue:
            if isinstance(Source, AudioSource):
                self.client.loop.run_coroutine_threadsafe(Source.cleanup)

    @property
    def crossfade(self) -> Union[float, None]:
        if self._gapless:
            return None

        return self._crossfade

    @crossfade.setter
    def crossfade(self, value: float) -> None:
        if self._gapless:
            raise ValueError("Cannot use crossfade when gapless mode is enabled.")

        self._crossfade = value

    @property
    def crossfadeVolume(self) -> Union[float, None]:
        if self._gapless:
            return None

        return 1.0 / (self.client.crossfade / Config.DELAY)

    @property
    def gapless(self) -> bool:
        return self._gapless

    @gapless.setter
    def gapless(self, value: bool) -> None:
        self._gapless = value

    def seek(self, offset: int) -> None:
        if not self.current:
            raise NotPlaying

        return self.current.seek(offset)

    @property
    def current(self) -> AudioSource:
        if not self._current:
            if self.next:
                self._current, self.next = self.next, None
            else:
                return None

        if self._current.filter != self.client.filter:
            self._current.filter = self.client.filter

        if not self._current.BufferLoader:
            self.client.dispatcher.dispatch("SOURCE_START", source=self._current)
            self._current.start()

        return self._current

    @current.setter
    def current(self, value: None) -> None:
        self.client.dispatcher.dispatch("SOURCE_END", source=self._current)

        self.client.loop.call_soon_threadsafe(self._current.cleanup)
        self._current = None

    @property
    def next(self) -> AudioSource:
        is_load_condition = (
            not self._current
            or self._current.stopped
            or self._current.remain <= (Config.PRELOAD_TIME + (self.crossfade or 0))
        )

        if not self._next:
            if not self.client.Queue:
                if self._current and is_load_condition and not self._request_dispatched:
                    self.client.dispatcher.dispatch(
                        "REQUIRE_NEXT_SOURCE", current=self._current
                    )
                    self._request_dispatched = True

                return None

            self._next = self.client.Queue[0]
            self._request_dispatched = False

        if isinstance(self._next, AudioData):

            def setSource(Source: AudioSource) -> None:
                if Source.AudioData == self.client.Queue[0]:
                    self._next = self.client.Queue[0] = Source

            self.getSource(self._next, setSource)

            return None

        if self._next != self.client.Queue[0]:
            self._next = None
            return None

        if self._next.filter != self.client.filter:
            self._next.filter = self.client.filter

        if self._current and not self._next.BufferLoader:
            if is_load_condition:
                self.client.dispatcher.dispatch("SOURCE_START", source=self._next)
                self._next.start()

        return self._next

    @next.setter
    def next(self, value: None) -> None:
        if not self._next:
            return

        if self.client.Queue:
            _nextItem = (
                self._next.AudioData
                if isinstance(self._next, AudioSource)
                else self._next
            )
            _queueItem = (
                self.client.Queue[0].AudioData
                if isinstance(self.client.Queue[0], AudioSource)
                else self.client.Queue[0]
            )

            if _nextItem == _queueItem:
                self.client.Queue.pop(0)
        self._next = None

    def getSource(self, *args, **kwargs) -> None:
        if not self._getSourceTask or self._getSourceTask.done():
            self._getSourceTask = asyncio.run_coroutine_threadsafe(
                self._getSource(*args, **kwargs), self.client.loop
            )

    async def _getSource(self, Data: AudioData, callback: Callable) -> None:
        Source = await Data.source()

        Source.volume = 1.0
        Source.filter = self.client.filter

        callback(Source)

    def read(self) -> bytes:
        if not self.current:
            return

        Data = self.current.read()

        if (
            not Data
            or self.current.volume <= 0.0
            or (self.gapless and self.current.stopped and self.next)
        ):
            self._crossfadeLoop = 0
            self.current = None

            return self.read()

        is_live = self.current.AudioData and self.current.AudioData.is_live
        is_crossfade_timing = (
            not self.gapless
            and self.next
            and (self.current.remain <= self.crossfade or self.current.stopped)
        )

        if not self.gapless and not is_live and is_crossfade_timing:
            NextData = self.next.read()
            if NextData:
                self._crossfadeLoop += 1
                crossfadeVolume = self.crossfadeVolume * self._crossfadeLoop

                self.current.volume = 1.0 - crossfadeVolume
                self.next.volume = crossfadeVolume

                Data = audioop.add(Data, NextData, 2)
        elif self.next:
            self.next.volume = 1.0

        if self.current.stopped:
            if self.current.volume > 0.0:
                self.current.volume = round(self.current.volume - 0.01, 3)
        elif not is_crossfade_timing:
            if self.current.volume < 1.0:
                self.current.volume = round(self.current.volume + 0.01, 3)
            elif self.current.volume > 1.0:
                self.current.volume = 1.0

        return Data

    def speak(self, state: bool) -> None:
        if self.client.speakState == state:
            return

        asyncio.run_coroutine_threadsafe(self.client.ws.speak(state), self.client.loop)

    def __do_run(self) -> None:
        self.loops = 0
        _start = time.perf_counter()

        while not self._end.is_set():
            try:
                if not self.client._connectedThread.is_set():
                    self.client._connectedThread.wait()
                    self.loops = 0
                    _start = time.perf_counter()

                Data = self.read() if not self.client.paused else None

                if self._volume != self.client.volume:
                    if self._volume < self.client.volume:
                        self._volume = round(self._volume + 0.01, 3)
                    if self._volume > self.client.volume:
                        self._volume = round(self._volume - 0.01, 3)

                self.speak(bool(Data))

                if Data:
                    if self._volume != 1.0:
                        Data = audioop.mul(Data, 2, min(self._volume, 2.0))

                    self.client.send(Data)

                self.loops += 1
                nextTime = _start + Config.DELAY * self.loops
                time.sleep(max(0, Config.DELAY + (nextTime - time.perf_counter())))
            except:
                traceback.print_exc()
                self.dispatcher.dispatch(
                    "PLAYER_TRACEBACK", traceback=traceback.format_exc()
                )

    def run(self) -> None:
        try:
            self.__do_run()
        except:
            pass
        finally:
            self.stop()

    def stop(self) -> None:
        self._end.set()
        self.speak(False)



import json
from typing import Coroutine

from sanic import Blueprint, response
from sanic.exceptions import abort

from ..config import Config
from ..source import AudioData

app = Blueprint(__name__)


def authorized(func: Coroutine) -> Coroutine:
    def wrapper(request, *args, **kwargs) -> Coroutine:
        if request.headers.get("Authorization") != Config.PASSWORD:
            abort(403, "Password mismatch.")

        return func(request, *args, **kwargs)

    return wrapper


def need_voiceclient(func: Coroutine) -> Coroutine:
    def wrapper(request, *args, **kwargs):
        user_id = int(request.headers.get("User-ID"))

        if (
            not hasattr(request.app, "ClientManagers")
            or not user_id in request.app.ClientManagers
        ):
            abort(404, "ClientManager not found.")

        manager = request.app.ClientManagers[user_id]

        guild_id = int(request.headers.get("Guild-ID"))

        VoiceClient = manager.getVC(guild_id, safe=True)

        VoiceClientID = request.headers.get("VoiceClient-ID")

        if not VoiceClient:
            abort(404, "VoiceClient not found.")

        if VoiceClientID != VoiceClient.id:
            abort(403, "VoiceClient ID mismtach.")

        return func(request, VoiceClient, *args, **kwargs)

    return wrapper


class Encoder(json.JSONEncoder):
    def default(self, o):
        return o.__dict__()


def JSONResponse(
    body,
    status=200,
    headers=None,
    content_type="application/json",
    **kwargs,
):
    return response.HTTPResponse(
        json.dumps(body, **kwargs, cls=Encoder),
        headers=headers,
        status=status,
        content_type=content_type,
    )


@app.get("/getSource")
@authorized
async def getSource(request) -> JSONResponse:
    Query = "".join(request.args.get("query", [])).strip()
    if not Query:
        abort(400, "Missing parameter query.")

    return JSONResponse({"source": await AudioData.create(Query)})


@app.post("/putSource")
@authorized
@need_voiceclient
async def putSource(request, VoiceClient) -> JSONResponse:
    if "source" not in request.json:
        abort(400, "Bad data `source`")

    index = VoiceClient.putSource(request.json["source"])

    return JSONResponse({"index": index})


@app.post("/loadSource")
@authorized
@need_voiceclient
async def loadSource(request, VoiceClient) -> JSONResponse:
    if "query" not in request.json:
        abort(400, "Bad data `query`")

    Source = await VoiceClient.loadSource(request.json["query"])

    return JSONResponse({"source": Source})


@app.post("/setVolume")
@authorized
@need_voiceclient
async def setVolume(request, VoiceClient) -> response.empty:
    if "volume" not in request.json or not isinstance(request.json["volume"], float):
        abort(400, "Bad data `volume`")

    VoiceClient.volume = request.json["volume"]

    return response.empty()


@app.post("/setCrossfade")
@authorized
@need_voiceclient
async def setCrossfade(request, VoiceClient) -> response.empty:
    if "crossfade" not in request.json or not isinstance(
        request.json["crossfade"], float
    ):
        abort(400, "Bad data `crossfade`")

    VoiceClient.crossfade = request.json["crossfade"]

    return response.empty()


@app.post("/setGapless")
@authorized
@need_voiceclient
async def setGapless(request, VoiceClient) -> response.empty:
    if "gapless" not in request.json or not isinstance(request.json["gapless"], bool):
        abort(400, "Bad data `gapless`")

    VoiceClient.gapless = request.json["gapless"]

    return response.empty()


@app.post("/setAutoplay")
@authorized
@need_voiceclient
async def setAutoplay(request, VoiceClient) -> response.empty:
    if "autoplay" not in request.json or not isinstance(request.json["autoplay"], bool):
        abort(400, "Bad data `autoplay`")

    VoiceClient.autoplay = request.json["autoplay"]

    return response.empty()


@app.post("/setFilter")
@authorized
@need_voiceclient
async def setFilter(request, VoiceClient) -> response.empty:
    if "filter" not in request.json or not isinstance(request.json["filter"], dict):
        abort(400, "Bad data `filter`")

    VoiceClient.filter = request.json["filter"]

    return response.empty()


@app.post("/seek")
@authorized
@need_voiceclient
async def seek(request, VoiceClient) -> response.empty:
    if "offset" not in request.json or not isinstance(request.json["offset"], float):
        abort(400, "Bad data `offset`")

    await VoiceClient.seek(request.json["offset"])

    return response.empty()


@app.post("/skip")
@authorized
@need_voiceclient
async def seek(request, VoiceClient) -> JSONResponse:
    if "offset" not in request.json or not isinstance(request.json["offset"], int):
        abort(400, "Bad data `offset`")

    VoiceClient.skip(request.json["offset"])

    return JSONResponse({"remain": len(VoiceClient.Queue)})


@app.post("/pause")
@authorized
@need_voiceclient
async def pause(request, VoiceClient) -> response.empty:
    VoiceClient.pause()

    return response.empty()


@app.post("/resume")
@authorized
@need_voiceclient
async def resume(request, VoiceClient) -> response.empty:
    VoiceClient.resume()

    return response.empty()


@app.post("/shuffle")
@authorized
@need_voiceclient
async def shuffle(request, VoiceClient) -> JSONResponse:
    VoiceClient.shuffle()

    return JSONResponse({"entries": VoiceClient.Queue})


@app.post("/remove")
@authorized
@need_voiceclient
async def remove(request, VoiceClient) -> JSONResponse:
    if "index" not in request.json or not isinstance(request.json["index"], int):
        abort(400, "Bad data `index`")

    removed = VoiceClient.Queue[request.json["index"] - 1]
    del VoiceClient.Queue[request.json["index"] - 1]

    return JSONResponse({"removed": removed, "entries": VoiceClient.Queue})


@app.get("/state")
@authorized
@need_voiceclient
async def resume(request, VoiceClient) -> JSONResponse:
    return JSONResponse(
        {
            "id": VoiceClient.id,
            "guild_id": VoiceClient.guild_id,
            "channel_id": VoiceClient.channel_id,
            "state": VoiceClient.state,
            "current": VoiceClient.current,
            "duration": VoiceClient.current.duration,
            "position": VoiceClient.current.position,
            "remain": VoiceClient.current.remain,
            "options": {
                "autoplay": VoiceClient.autoplay,
                "volume": VoiceClient.volume,
                "crossfade": VoiceClient.crossfade,
                "gapless": VoiceClient.gapless,
                "filter": VoiceClient.filter,
            },
        }
    )


@app.get("/queue")
@authorized
@need_voiceclient
async def queue(request, VoiceClient) -> JSONResponse:
    return JSONResponse({"entries": VoiceClient.Queue})
